#include <iostream>
#include <fstream>
#include <vector>

typedef unsigned int uint;

struct sparseMatrix {

  uint size; // Size of whole matrix
  uint nNz;  // # of non zero terms

  std::vector<uint  > rows,cols; // row and column indices of non zero terms
  std::vector<double> vals;      // non zero terms

};

// Reads coefficients of a sparse matrix from a text file
// The file contains the number of non zero terms on its first line
// then on each line the row index, column index and value of non zero term
sparseMatrix readMatrix(std::string fileName) {

  sparseMatrix M;
  std::ifstream mFile(fileName.c_str());
  
  M.size = 0;

  mFile >> M.nNz;
  M.rows.resize(M.nNz);
  M.cols.resize(M.nNz);
  M.vals.resize(M.nNz);
  
  for (uint i=0u; i<M.nNz; i++) {
    mFile >> M.rows[i] >> M.cols[i] >> M.vals[i];
    if (M.rows[i] > M.size) M.size = M.rows[i];
    if (M.cols[i] > M.size) M.size = M.cols[i];
  }

  M.size++; // Indices start at 0!
  return M;

}

// Prints on screen the contents of a sparse matrix
void displayMatrix(const sparseMatrix& M) {
  std::cout << "Sparse matrix: size: " << M.size << ", non-zero terms (" << M.nNz << "):" << std::endl;
  for (uint i=0u; i<M.nNz; i++)
    std::cout << "(" << M.rows[i] << "," << M.cols[i] << ") : " << M.vals[i] << std::endl;
}

// Computes trace of matrix (i.e. sum of diagonal terms)
double trace(const sparseMatrix& M) {
  double res = 0.e0;
  for (uint i=0u; i<M.nNz; i++)
    if (M.rows[i]==M.cols[i])
      res += M.vals[i];
  return res;
}

// Computes infinite norm of matrix (max over rows of sum of abs of coefficients)
double inftNorm(const sparseMatrix& M) {
  if (M.nNz==0) {
    std::cout << "Warning: empty matrix !" << std::endl;
    return 0.;
  } 
 
  // Compute sums
  std::vector<double> maxs(M.size,0.);
  for (uint i=0u; i<M.nNz; i++)
    maxs[M.rows[i]] += abs(M.vals[i]);

  // Get max of sums
  double res = maxs[0];
  for (uint i=1u; i<M.size; i++)
    if (maxs[i]>res) res = maxs[i];
  return res;
}

// Fills res with the matrix vector product Mx.
void matMul(const sparseMatrix& M, const std::vector<double>& x, std::vector<double>& res) {

  // Check matrix and vector shapes
  if (x.size()>M.size)
    std::cout << "matMul: Warning, will only use " << M.size << " first coefficients of x (size of matrix)." << std::endl;
  else if (x.size()<M.size) {
    std::cerr << "matMul: Vector (" << x.size() << ") is smaller than matrix (" << M.size << ")." << std::endl;
    abort();
  }

  // (re)initialize result
  res.clear();
  for (uint i=0u; i<x.size(); i++)
    res.push_back(0.);

  // Add each matrix coefficient contribution
  for (uint i=0u; i<M.nNz; i++)
    res[M.rows[i]] += M.vals[i]*x[M.cols[i]];

}

double inftNorm(const std::vector<double>& x) {
  if (x.size()==0) {
    std::cout << "Warning: empty vector !" << std::endl;
    return 0.;
  }  
  double res = abs(x[0]);
  for (uint i=1u; i<x.size(); i++)
    if (abs(x[i])>res) 
      res = abs(x[i]);
  return res;
}

// Uses power iteration method to compute dominating eigen value of sparse matrix
double powerIteration(const sparseMatrix& M, int maxIters = 100, double tolerance = 1.e-4) {

  // Initialize algorithm
  std::vector<double> bk  (M.size,0.);
  std::vector<double> Mbk (M.size,0.);
  std::vector<double> bkP1(M.size,0.);
  for (double& b: bk)
    b = std::rand()/(double)RAND_MAX;

  matMul(M,bk,Mbk);
  double normk = inftNorm(Mbk), normkP1;
  double diff  = 1.e3;
  int    iter  = 0;

  // Iterate
  do {
    // Compute bk+1 = Mbk / norm(Mbk)
    for (uint i=0u; i<M.size; i++)
      bkP1[i] = Mbk[i] / normk;    
    
    // Compute MbK+1 and its norm (we put the result in Mbk)
    matMul(M,bkP1,Mbk);
    normkP1 = inftNorm(Mbk);
    
    // Determine if we have to stop (relative difference of result)
    diff = abs(normkP1-normk)/normk;
    iter++;

    // Prepare next iteration
    normk = normkP1;


  } while (diff > tolerance and iter < maxIters); 
 
  std::cout << "Power iteration stopped after " << iter << " iterations." << std::endl; 

  return normk;
}

int main() {

  sparseMatrix A = readMatrix("matrix.dat"); 
  displayMatrix(A); 

  std::cout << "Trace of read matrix is " << trace(A) << std::endl;
  std::cout << "Its infinite norm is " << inftNorm(A) << std::endl;
 
  /*
  std::cout << "Test matMul " << std::endl;
  std::vector<double> x(A.size,0.);
  std::vector<double> r(A.size,0.);
  for (uint i=0u; i<x.size(); i++) {
    x[i] = 1.;
    matMul(A,x,r);
    std::cout << i << " r ";
    for (uint j=0; j<r.size(); j++)
      std::cout << r[j] << " ";
    std::cout << std::endl;
    x[i] = 0.;
  }
  */

  double eigenValue = powerIteration(A);
  std::cout << "Largest eigen value is " << eigenValue << std::endl;

}



